from typing import Optional, Any, Type, Union, Dict, List
import httpx
from pydantic import BaseModel, ValidationError
from tenacity import RetryError
from aiolimiter import AsyncLimiter
import asyncio
from collections import deque
import time

from .auth import AuthStrategy, NoAuth
from .exceptions import RateLimitError, RateLimitFailure, TaskAborted
from .response import BaseAPIResponse
from .utils import paginate_requests

import logging
import structlog

# Configure logging
_logger = logging.getLogger(__name__)
log = structlog.wrap_logger(_logger)


class APIClient:
    """Base class for API client"""
    is_failed: bool = False

    base_url: str
    timeout: Optional[float] = 10.0  # Default timeout of 10 seconds
    verify_ssl: bool = True
    requests_per_second: int = None
    requests_per_minute: int = None
    verbose: bool = False

    auth_strategy: Optional[AuthStrategy] = None
    response_class: Optional[Type[BaseAPIResponse]] = None
    response_model: Optional[Type[BaseModel]] = None
    limiter_per_second: Optional[AsyncLimiter] = None
    limiter_per_minute: Optional[AsyncLimiter] = None
    _request_timestamps: deque = deque()
    __max_requests_per_second: int = 0
    __total_requests: int = 0
    __total_retried_requests: int = 0

    # New attributes for retry configuration
    max_retries: int = 0  # Default to no retries
    max_retry_wait: float = 5.0  # Maximum wait time between retries in seconds

    def __init__(self,
                 base_url: str,
                 *,  # Force key-value pairs for input
                 auth_strategy: Optional[AuthStrategy] = None,
                 response_class: Optional[Type[BaseAPIResponse]] = None,
                 response_model: Optional[Type[BaseModel]] = None,
                 timeout: float = 10.0,
                 default_params: dict = None,
                 default_headers: dict = None,
                 verify_ssl: bool = True,
                 requests_per_minute: int = None,
                 requests_per_second: int = None,
                 verbose: bool = False,
                 max_retries: int = 0,
                 max_retry_wait: float = 5.0,
                 httpx_kwargs: dict = None,
                 ):
        self.base_url = base_url
        self.timeout = timeout
        self.verify_ssl = verify_ssl
        self.requests_per_minute = requests_per_minute
        self.requests_per_second = requests_per_second
        self.verbose = verbose
        self.auth_strategy = auth_strategy or NoAuth()
        self.response_class = response_class or BaseAPIResponse
        self.response_model = response_model
        self.max_retries = max_retries
        self.max_retry_wait = max_retry_wait
        # Initialize client here
        client_params = dict(base_url=self.base_url, timeout=self.timeout, verify=self.verify_ssl)
        if default_params:
            client_params["params"] = default_params
        if default_headers:
            client_params["headers"] = default_headers
        if httpx_kwargs:
            client_params.update(httpx_kwargs)
        self.client = httpx.AsyncClient(**client_params)

        # Initialize rate limiters
        if self.requests_per_second:
            self.limiter_per_second = AsyncLimiter(self.requests_per_second, time_period=1)
        if self.requests_per_minute:
            self.limiter_per_minute = AsyncLimiter(self.requests_per_minute, time_period=60)

    async def __aenter__(self) -> 'APIClient':
        # No need to initialize the client here as it is already initialized in __init__
        return self

    async def __aexit__(self, exc_type: Optional[Type[BaseException]], exc_value: Optional[BaseException], traceback: Optional[Any]):
        if self.client:
            await self.client.aclose()

    @property
    def max_requests_per_second(self) -> int:
        """Returns the highest rate of requests per second reached."""
        return self.__max_requests_per_second

    @property
    def total_requests(self) -> int:
        """Returns the highest rate of requests per second reached."""
        return self.__total_requests

    @property
    def total_retried_requests(self):
        return self.__total_retried_requests

    def log_verbose(self, msg, logger=None, **kwargs):
        if not logger:
            logger = log
        if self.verbose:
            logger.debug(msg, **kwargs)

    async def create_response_object(self, response: httpx.Response, response_class: Optional[Type[BaseAPIResponse]]):
        response_class = response_class or self.response_class
        response_obj = response_class(api_client=self, response=response, response_model=self.response_model)
        if response_obj.content is None:
            await response_obj._async_parse_content()
        return response_obj

    async def _perform_request(self, method: str, endpoint: str, **kwargs) -> httpx.Response:
        """Helper function to send the HTTP request."""
        request = self.client.build_request(method=method, url=endpoint, **kwargs)
        self.auth_strategy.authenticate(request)
        return await self.client.send(request)

    async def _handle_response(self, response: httpx.Response, request: httpx.Request, response_class: Optional[Type[BaseAPIResponse]]) -> BaseAPIResponse:
        """Process the response and handle errors."""
        response_obj = await self.create_response_object(response=response, response_class=response_class)
        if response_obj.is_rate_limit_error:
            raise RateLimitError("Rate limit error", request=request, response=response)
        if response_obj.is_rate_limit_failure:
            raise RateLimitFailure("Rate limit failure", request=request, response=response)
        return response_obj

    @paginate_requests
    async def send_request(self, method: str, endpoint: str, raise_for_status=True, result_limit: int = None, response_class: Optional[Type[BaseAPIResponse]] = None, **kwargs) -> Union[httpx.Response, BaseAPIResponse, Dict, List, Any]:
        """Send an HTTP request with retries and authentication"""
        _ = result_limit  # Used only by pagination
        __logger = log.new(method=method, url=endpoint)
        if self.is_failed:
            __logger.fatal("Failure reported; aborting tasks")
            raise TaskAborted("Failure reported; aborting tasks")
        __logger.debug("Sending request")
        if self.client is None:
            raise RuntimeError("HTTP client is not initialized")

        self.__total_requests += 1
        retries = 0
        while retries <= self.max_retries:
            try:
                # Apply both rate limiters
                if self.limiter_per_second and self.limiter_per_minute:
                    async with self.limiter_per_second, self.limiter_per_minute:
                        response = await self._perform_request(method, endpoint, **kwargs)
                elif self.limiter_per_second:
                    async with self.limiter_per_second:
                        response = await self._perform_request(method, endpoint, **kwargs)
                elif self.limiter_per_minute:
                    async with self.limiter_per_minute:
                        response = await self._perform_request(method, endpoint, **kwargs)
                else:
                    response = await self._perform_request(method, endpoint, **kwargs)

                # Log the request time
                current_time = time.time()
                self._request_timestamps.append(current_time)
                # Remove timestamps older than 1 second
                while self._request_timestamps and self._request_timestamps[0] < current_time - 1:
                    self._request_timestamps.popleft()
                # Update the max requests per second
                current_rate = len(self._request_timestamps)
                if current_rate > self.__max_requests_per_second:
                    self.__max_requests_per_second = current_rate

                self.log_verbose("Received response", status_code=response.status_code, logger=__logger)
                response_obj = await self._handle_response(response=response, request=response.request, response_class=response_class)
                if raise_for_status:
                    response.raise_for_status()
                return response_obj
            except httpx.TransportError as e:  # Retry only on connection errors for now
                retries += 1
                if retries > self.max_retries:
                    if self.max_retries > 0:
                        __logger.error(f"Exceeded maximum retries ({self.max_retries})")
                    raise
                self.__total_retried_requests += 1
                __logger.warning(f"Connection error occurred: {e}. Retry {retries}/{self.max_retries}.")
                # Exponential backoff with a max wait time
                retry_wait = min(2 ** retries, self.max_retry_wait)
                await asyncio.sleep(retry_wait)
            except httpx.HTTPStatusError as e:
                __logger.debug("[ERROR] HTTP error occurred", status_code=e.response.status_code, content=e.response.text)
                raise
            except RetryError as e:
                __logger.debug("[ERROR] Retry failed", error=str(e))
                raise
            except ValidationError as e:
                __logger.debug("[ERROR] Response validation failed", error=str(e), method=method, url=endpoint)
                raise

    async def get(self, endpoint: str, **kwargs):
        """Send a GET request"""
        return await self.send_request(method="GET", endpoint=endpoint, **kwargs)

    async def post(self, endpoint: str, **kwargs):
        """Send a POST request"""
        return await self.send_request(method="POST", endpoint=endpoint, **kwargs)



from tenacity import retry, stop_after_attempt, wait_exponential
from .response import APIResponse

import logging
import structlog

_logger = logging.getLogger(__name__)
log = structlog.wrap_logger(_logger)


# ToDo Revisit this once I decide what I really want to offer with respect to retries
def with_retries(func):
    """Decorator to add retries to a function based on config"""
    def wrapper(self, *args, **kwargs):
        retry_decorator = retry(stop=stop_after_attempt(self.config.retry_attempts), wait=wait_exponential(min=1, max=10))
        decorated_func = retry_decorator(func)
        return decorated_func(self, *args, **kwargs)
    return wrapper


# ToDo In its current form, if the initial request contains a parameter like what page to start on, subsequent calls based on URL are still overwritten by that original param
#  This works as long as the request starts at the beginning, but if user requests from a specific starting point, it will just keep fetching that same list forever
def paginate_requests(func):
    """Decorator to handle pagination in API responses"""
    async def wrapper(self, *args, **kwargs):
        response_obj: APIResponse = await func(self, *args, **kwargs)
        if not response_obj.is_pagination:
            return response_obj
        log.debug("Pagination detected", url=str(response_obj.response.url))
        result_limit = kwargs.pop("result_limit", 0)
        all_responses = [response_obj]
        all_results = [r for r in response_obj.result_list or []]
        urls_fetched = []
        pagination_payload = await response_obj.get_pagination_payload()
        while pagination_payload:
            if result_limit and len(all_results) >= result_limit:
                break
            kwargs.update(pagination_payload)
            next_url = kwargs.get("endpoint")
            if next_url in urls_fetched:
                raise Exception(f"Next URL is the same as one previously requested already. URL: {next_url}")
            if response_obj.response.request.method == "GET":
                urls_fetched.append(next_url)
            log.debug("Continuing pagination", url=next_url)
            response_obj = await func(self, *args, **kwargs)
            all_responses.append(response_obj)
            if response_obj.result_list:
                all_results.extend(response_obj.result_list)
            pagination_payload = await response_obj.get_pagination_payload()

        if result_limit:
            return all_responses, all_results[:result_limit]
        return all_responses, all_results
    return wrapper



from WedgieIntegrator.config import APIConfig

def test_api_config_defaults():
    config = APIConfig(base_url="https://example.com")
    assert config.base_url == "https://example.com"
    # assert config.retry_attempts == 3
    assert config.timeout == 10.0
    assert config.verify_ssl is True

def test_api_config_custom():
    config = APIConfig(base_url="https://example.com", timeout=20.0, verify_ssl=False)
    assert config.base_url == "https://example.com"
    # assert config.retry_attempts == 5
    assert config.timeout == 20.0
    assert config.verify_ssl is False



import pytest
from WedgieIntegrator.utils import with_retries
from tenacity import RetryError

# ToDo Revisit this once I decide what I really want to offer with respect to retries

class TestClient:
    def __init__(self):
        self.config = type('Config', (), {'retry_attempts': 3})

    @with_retries
    def unreliable_method(self):
        raise Exception("Temporary failure")

    @with_retries
    def reliable_method(self):
        return "Success"

def test_with_retries_success():
    client = TestClient()
    result = client.reliable_method()
    assert result == "Success"

def test_with_retries_failure():
    client = TestClient()
    with pytest.raises(RetryError):
        client.unreliable_method()



# Workarounds for asyncio in Python 3.7

import asyncio
from concurrent.futures import ThreadPoolExecutor

executor = ThreadPoolExecutor()


# Workaround for Python 3.7 because asyncio.run was added in 3.8
def run(coro):
    try:
        # Check if there's a running event loop
        loop = asyncio.get_event_loop()
        if loop.is_running():
            # If there's a running loop, we can't use run_until_complete
            return asyncio.run_coroutine_threadsafe(coro, loop).result()

        # If the loop is not running, we can use it
        return loop.run_until_complete(coro)
    except RuntimeError:
        # Create a new event loop if none exists or if the loop is closed
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        return loop.run_until_complete(coro)
    finally:
        # Do not close the loop if it's running or if it was newly created in this call
        if not loop.is_running():
            loop.close()


# Workaround for Python 3.7 because asyncio.run was added in 3.9
async def to_thread(func, *args, **kwargs):
    loop = asyncio.get_event_loop()
    return await loop.run_in_executor(executor, func, *args, **kwargs)



from httpx import HTTPStatusError


class BaseClientException(Exception):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class RateLimitError(HTTPStatusError, BaseClientException): ...


class RateLimitFailure(HTTPStatusError, BaseClientException): ...


class TaskAborted(BaseClientException): ...



from typing import Optional, Any, Type, Union

import httpx
from pydantic import BaseModel
try:
    from asyncio import to_thread
except ImportError:
    from .asyncio_workaround import to_thread


class BaseAPIResponse:
    response: httpx.Response
    response_model: Optional[Type[BaseModel]] = None
    content_type: str
    is_rate_limit_error: bool = False
    is_rate_limit_failure: bool = False
    is_pagination: bool = False
    link_header: str = None
    pagination_links: dict = None
    __content = None
    __client = None

    def __init__(self, api_client, response: httpx.Response, response_model: Optional[Type[BaseModel]] = None):
        self.__client = api_client
        self.response = response
        self.response_model = response_model
        self.content_type = response.headers.get('Content-Type', '')
        if self.response.status_code == 429:
            self.is_rate_limit_error = True

    @property
    def content(self) -> Union[dict, list, Any]:
        # Remember that this is not accessible until after initialization, because _async_parse_content has to run first
        return self.__content

    @property
    def result_list(self):
        """Customizable property for returning results as a list, when applicable"""
        if isinstance(self.content, list):
            return self.content

    async def is_json(self):
        """Standalone parser to make customization easy"""
        if 'application/json' in self.content_type:
            return True
        return False

    @property
    def pagination_next_link(self):
        if self.pagination_links:
            return self.pagination_links.get('next')

    async def get_pagination_payload(self):
        request_args = {}
        if self.pagination_next_link:
            request_args['endpoint'] = self.pagination_next_link
        return request_args

    async def _async_parse_content(self):
        if self.response_model:
            parsed_response = await to_thread(self.response.json)
            self.__content = self.response_model.parse_obj(parsed_response)
        elif await self.is_json():
            self.__content = await to_thread(self.response.json)
        elif 'text/' in self.content_type:
            self.__content = self.response.text
        else:
            self.__content = self.response.content


class APIResponse(BaseAPIResponse):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.link_header = self.response.headers.get('Link')
        if self.link_header:
            self.pagination_links = {}
            for link in self.link_header.split(','):
                parts = link.split(';')
                url = parts[0].strip('<> ')
                rel = parts[1].strip().split('=')[1].strip('"')
                self.pagination_links[rel] = url
            if self.response.request.method == "GET":
                self.is_pagination = True



"""
For testing WedgieIntegrator with a no auth API
"""

import pytest
from WedgieIntegrator.config import APIConfig
from WedgieIntegrator.client import BaseAPIClient
from WedgieIntegrator.auth import NoAuth

# ToDo this was supposed to just test this package with a no auth API, but it turned into more comprehensive testing.
#  Revisit and rethink...

@pytest.fixture
def api_config():
    return APIConfig(base_url="https://jsonplaceholder.typicode.com", api_key=None, oauth_token=None)

@pytest.fixture
def api_client(api_config):
    auth_strategy = NoAuth()
    return BaseAPIClient(config=api_config, auth_strategy=auth_strategy)

@pytest.mark.asyncio
async def test_get_all_posts(api_client):
    """Test GET request to retrieve all posts"""
    async with api_client:
        response_obj = await api_client.get(endpoint="/posts")
        assert isinstance(response_obj.content, list)
        assert len(response_obj.content) > 0
        assert 'id' in response_obj.content[0]

@pytest.mark.asyncio
async def test_get_single_post(api_client):
    """Test GET request to retrieve a single post"""
    async with api_client:
        response_obj = await api_client.get(endpoint="/posts/1")
        assert isinstance(response_obj.content, dict)
        assert response_obj.content['id'] == 1

@pytest.mark.asyncio
async def test_create_post(api_client):
    """Test POST request to create a new post"""
    post_data = {
        'title': 'foo',
        'body': 'bar',
        'userId': 1
    }
    async with api_client:
        response_obj = await api_client.post(endpoint="/posts", json=post_data)
        assert isinstance(response_obj.content, dict)
        assert response_obj.content['title'] == 'foo'
        assert response_obj.content['body'] == 'bar'
        assert response_obj.content['userId'] == 1

@pytest.mark.asyncio
async def test_update_post(api_client):
    """Test PUT request to update an existing post"""
    update_data = {
        'id': 1,
        'title': 'foo',
        'body': 'bar',
        'userId': 1
    }
    async with api_client:
        response_obj = await api_client.send_request(method="PUT", endpoint="/posts/1", json=update_data)
        assert isinstance(response_obj.content, dict)
        assert response_obj.content['title'] == 'foo'
        assert response_obj.content['body'] == 'bar'

@pytest.mark.asyncio
async def test_delete_post(api_client):
    """Test DELETE request to delete a post"""
    async with api_client:
        response_obj = await api_client.send_request(method="DELETE", endpoint="/posts/1")
        assert response_obj.content == {}



import pytest
from WedgieIntegrator.auth import BearerTokenAuth, BasicAuth, NoAuth, TokenAuth
from httpx import Request
import base64

def test_no_auth():
    auth = NoAuth()
    request = Request(method="GET", url="https://example.com")
    auth.authenticate(request)
    assert "Authorization" not in request.headers

def test_basic_auth():
    auth = BasicAuth(username="dummy_user", password="dummy_pass")
    request = Request(method="GET", url="https://example.com")
    auth.authenticate(request)
    expected_auth_header = "Basic " + base64.b64encode(b"dummy_user:dummy_pass").decode("utf-8")
    assert request.headers["Authorization"] == expected_auth_header

# @pytest.fixture
# def request():
#     return Request(method="GET", url="https://example.com")
#
# @pytest.mark.parametrize("auth_class, token, expected_header", [
#     (APIKeyAuth, "dummy_api_key", "Bearer dummy_api_key"),
#     (OAuthAuth, "dummy_oauth_token", "Bearer dummy_oauth_token"),
#     (BearerTokenAuth, "dummy_bearer_token", "Bearer dummy_bearer_token"),
# ])

def test_api_key_auth():
    auth = TokenAuth(token="dummy_api_key")
    request = Request(method="GET", url="https://example.com")
    auth.authenticate(request)
    assert request.headers["Authorization"] == "Bearer dummy_api_key"

# def test_oauth_auth():
#     auth = OAuthAuth(token="dummy_oauth_token")
#     request = Request(method="GET", url="https://example.com")
#     auth.authenticate(request)
#     assert request.headers["Authorization"] == "Bearer dummy_oauth_token"

def test_bearer_token_auth():
    auth = BearerTokenAuth(token="dummy_bearer_token")
    request = Request(method="GET", url="https://example.com")
    auth.authenticate(request)
    assert request.headers["Authorization"] == "Bearer dummy_bearer_token"



import pytest
import httpx
from WedgieIntegrator.client import APIClient
from WedgieIntegrator.config import APIConfig
from WedgieIntegrator.auth import BasicAuth, BearerTokenAuth, TokenAuth
from httpx import Request, Response
from unittest.mock import patch

class MockAuth(TokenAuth):
    """Mock authentication strategy for testing"""
    def authenticate(self, request: Request) -> None:
        request.headers['Authorization'] = 'MockAuth'

@pytest.fixture
def api_config():
    return APIConfig(base_url="https://api.example.com")

@pytest.fixture
def api_client(api_config):
    auth_strategy = MockAuth(token="dummy_api_key")
    return APIClient(config=api_config, auth_strategy=auth_strategy)

def test_api_config(api_config):
    """Test APIConfig initialization"""
    assert api_config.base_url == "https://api.example.com"
    # assert api_config.retry_attempts == 3
    assert api_config.timeout == 10.0
    assert api_config.verify_ssl == True

def test_auth_with_api_key():
    """Test APIKeyAuth authentication"""
    auth = TokenAuth(token="dummy_api_key")
    request = Request(method="GET", url="https://api.example.com")
    auth.authenticate(request)
    assert request.headers["Authorization"] == "Bearer dummy_api_key"

# def test_auth_with_oauth():
#     """Test OAuthAuth authentication"""
#     auth = OAuthAuth(token="dummy_oauth_token")
#     request = Request(method="GET", url="https://api.example.com")
#     auth.authenticate(request)
#     assert request.headers["Authorization"] == "Bearer dummy_oauth_token"

def test_auth_with_bearer_token():
    """Test BearerTokenAuth authentication"""
    auth = BearerTokenAuth(token="dummy_bearer_token")
    request = Request(method="GET", url="https://api.example.com")
    auth.authenticate(request)
    assert request.headers["Authorization"] == "Bearer dummy_bearer_token"

def test_auth_with_basic_auth():
    """Test BasicAuth authentication"""
    auth = BasicAuth(username="dummy_user", password="dummy_pass")
    request = Request(method="GET", url="https://api.example.com")
    auth.authenticate(request)
    assert request.headers["Authorization"].startswith("Basic ")

@patch.object(httpx.AsyncClient, 'send', return_value=httpx.Response(200, request=Request("GET", "https://api.example.com"), json={"key": "value"}))
@pytest.mark.asyncio
async def test_send_request(mock_send, api_client):
    """Test sending an HTTP request"""
    response_obj = await api_client.send_request(method="GET", endpoint="/test")
    assert response_obj.content == {"key": "value"}
    # ToDo revisit
    # assert mock_send.call_args[1]["method"] == "GET"

@patch.object(httpx.AsyncClient, 'send', return_value=httpx.Response(200, request=Request("GET", "https://api.example.com"), json={"key": "value"}))
@pytest.mark.asyncio
async def test_get(mock_send, api_client):
    """Test sending a GET request"""
    response_obj = await api_client.get(endpoint="/test")
    assert response_obj.content == {"key": "value"}
    # ToDo revisit
    # assert mock_send.call_args[1]["method"] == "GET"

@patch.object(httpx.AsyncClient, 'send', return_value=httpx.Response(200, request=Request("POST", "https://api.example.com"), json={"key": "value"}))
@pytest.mark.asyncio
async def test_post(mock_send, api_client):
    """Test sending a POST request"""
    response_obj = await api_client.post(endpoint="/test")
    assert response_obj.content == {"key": "value"}
    # ToDo revisit
    # assert mock_send.call_args[1]["method"] == "POST"

@pytest.mark.parametrize("auth_class, token, header_value", [
    # (OAuthAuth, "dummy_oauth_token", "Bearer dummy_oauth_token"),
    (BearerTokenAuth, "dummy_bearer_token", "Bearer dummy_bearer_token"),
    (TokenAuth, "dummy_api_key", "Bearer dummy_api_key"),
])
def test_auth(auth_class, token, header_value):
    """Test different authentication strategies"""
    auth = auth_class(token)
    request = Request(method="GET", url="https://api.example.com")
    auth.authenticate(request)
    assert request.headers["Authorization"] == header_value

@patch.object(httpx.AsyncClient, 'send', side_effect=httpx.HTTPStatusError("Error", request=Request("GET", "https://api.example.com"), response=Response(400)))
@pytest.mark.asyncio
async def test_send_request_http_error(mock_send, api_client):
    """Test handling HTTP status errors in send_request"""
    with pytest.raises(httpx.HTTPStatusError):
        await api_client.send_request(method="GET", endpoint="/test")



import base64
from abc import ABC, abstractmethod
from dataclasses import dataclass

from httpx import Request


class AuthStrategy(ABC):
    """Abstract base class for authentication strategies"""

    @abstractmethod
    def authenticate(self, request: Request):
        """Apply authentication to the request"""
        pass


class NoAuth(AuthStrategy):
    """No authentication strategy."""

    def authenticate(self, request: Request):
        pass


@dataclass
class HeaderAuth(AuthStrategy):
    """Generic token authentication strategy"""
    header_name: str = "Authorization"
    header_prefix: str = None

    def __init__(self, secret: str):
        self.__secret = secret

    def authenticate(self, request: Request):
        if not self.__secret:
            return
        if self.header_prefix:
            request.headers['Authorization'] = f"{self.header_prefix} {self.__secret}"
        else:
            request.headers['Authorization'] = self.__secret


class BasicAuth(HeaderAuth):
    """Basic authentication strategy"""
    header_prefix = "Basic"

    def __init__(self, username, password, header_name: str = None, header_prefix: str = None):
        if header_name is not None:
            self.header_name = header_name
        if header_prefix is not None:
            self.header_prefix = header_prefix
        basic_auth_str = f"{username}:{password}"
        super().__init__(secret=base64.b64encode(basic_auth_str.encode("utf-8")).decode("utf-8"))


class TokenAuth(HeaderAuth):
    """Generic token authentication strategy"""
    header_prefix: str = "Bearer"

    def __init__(self, token: str, header_name: str = None, header_prefix: str = None):
        if header_name is not None:
            self.header_name = header_name
        if header_prefix is not None:
            self.header_prefix = header_prefix
        super().__init__(secret=token)


class BearerTokenAuth(TokenAuth):
    """Bearer token authentication strategy"""

    def __init__(self, token: str):
        super().__init__(token=token)


# @dataclass
# class OAuthAuth(TokenAuth):
#     """OAuth token authentication strategy"""
#     token: str
#     header_name: str = "Authorization"
#     header_prefix: str = "Bearer"



import os
import pytest
from WedgieIntegrator.config import APIConfig
from WedgieIntegrator.client import BaseAPIClient
from WedgieIntegrator.auth import BasicAuth

GITHUB_USERNAME = os.getenv('GITHUB_USERNAME')
GITHUB_PASSWORD = os.getenv('GITHUB_PASSWORD')
TEST_REPO_NAME = 'test-repo'

@pytest.fixture
def api_config():
    return APIConfig(base_url="https://api.github.com")

@pytest.fixture
def api_client(api_config):
    if not GITHUB_USERNAME or not GITHUB_PASSWORD:
        pytest.fail("Environment variables GITHUB_USERNAME and GITHUB_PASSWORD must be set")
    auth_strategy = BasicAuth(username=GITHUB_USERNAME, password=GITHUB_PASSWORD)
    return BaseAPIClient(config=api_config, auth_strategy=auth_strategy)

@pytest.mark.asyncio
async def test_get_authenticated_user(api_client):
    """Test GET request to retrieve authenticated user's details"""
    async with api_client:
        response_obj = await api_client.get(endpoint="/user")
        assert isinstance(response_obj.content, dict)
        assert 'login' in response_obj.content

@pytest.mark.asyncio
async def test_get_user_repos(api_client):
    """Test GET request to retrieve authenticated user's repositories"""
    async with api_client:
        response_obj = await api_client.get(endpoint="/user/repos")
        assert isinstance(response_obj.content, list)
        if response_obj.content:
            assert 'name' in response_obj.content[0]

@pytest.mark.asyncio
async def test_create_repo(api_client):
    """Test POST request to create a new repository"""
    repo_data = {
        'name': TEST_REPO_NAME,
        'description': 'This is a test repository',
        'private': False
    }
    async with api_client:
        response_obj = await api_client.post(endpoint="/user/repos", json=repo_data)
        assert isinstance(response_obj.content, dict)
        assert response_obj.content['name'] == TEST_REPO_NAME
        assert response_obj.content['description'] == 'This is a test repository'

@pytest.mark.asyncio
async def test_update_repo(api_client):
    """Test PATCH request to update an existing repository"""
    update_data = {
        'name': TEST_REPO_NAME,
        'description': 'This is an updated test repository',
        'private': False
    }
    async with api_client:
        response_obj = await api_client.send_request(method="PATCH", endpoint=f"/repos/{GITHUB_USERNAME}/{TEST_REPO_NAME}", json=update_data)
        assert isinstance(response_obj.content, dict)
        assert response_obj.content['description'] == 'This is an updated test repository'

@pytest.mark.asyncio
async def test_delete_repo(api_client):
    """Test DELETE request to delete a repository"""
    async with api_client:
        response_obj = await api_client.send_request(method="DELETE", endpoint=f"/repos/{GITHUB_USERNAME}/{TEST_REPO_NAME}")
        assert response_obj.response.status_code == 204  # No content



"""
For testing WedgieIntegrator with a no auth API
"""

import pytest
from WedgieIntegrator.config import APIConfig
from WedgieIntegrator.client import BaseAPIClient
from WedgieIntegrator.auth import NoAuth

@pytest.fixture
def api_config():
    # return APIConfig(base_url="https://jsonplaceholder.typicode.com", api_key=None, oauth_token=None)
    return APIConfig(base_url="https://httpbin.org", api_key=None, oauth_token=None)


@pytest.fixture
def api_client(api_config):
    auth_strategy = NoAuth()
    return BaseAPIClient(config=api_config, auth_strategy=auth_strategy)


@pytest.mark.asyncio
async def test_status_code_200(api_client):
    """Test GET request to retrieve a single post"""
    async with api_client:
        status_code = 200
        print(f"PRINTING /status/{status_code}")
        response = await api_client.get(endpoint=f"/status/{status_code}")
        print(type(response))
        assert isinstance(response, str)
        # assert isinstance(response, dict)



from setuptools import setup, find_packages

setup(
    name="WedgieIntegrator",
    version="0.1.2.14",
    author="Chad Roberts",
    author_email="jcbroberts@gmail.com",
    description="An API client toolkit that is async friendly",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/deathbywedgie/WedgieIntegrator",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    install_requires=[
        "httpx",
        "pydantic",
        "tenacity",
        "structlog",
        "aiolimiter",
    ],
)
